///////////////////////////////////////////////////////////////////////
// Code for user modes
///////////////////////////////////////////////////////////////////////

void MainMode_0() {
  ///////////////////////////////////////////////////////////////////////
  // In waiting for user input mode (default)
  ///////////////////////////////////////////////////////////////////////
  // slow circular LED pattern, waiting for button press
}

// --------------------------------------------------------------------------------

void MainMode_1() {
  ///////////////////////////////////////////////////////////////////////
  // flash LEDs to simulate spirit level to determine upright position
  ///////////////////////////////////////////////////////////////////////
  // extinguish all LEDs
  ClrLEDFlags();
  // calculate X+Y acceleration vector and angle
  if (accelerometer_data_rawY_OA == 0) {accelerometer_data_rawY_OA = 1;} // avoid divide by zero
  AccAbsAng = abs((float)accelerometer_data_rawX_OA/(float)accelerometer_data_rawY_OA);
  AccVect = sqrt(sq((float)accelerometer_data_rawX_OA) + sq((float)accelerometer_data_rawY_OA));
  AccVectTrk += ((AccVect - AccVectTrk)/10.0); // apply noise filtering
  ///////////////////////////////////////////////////////////////////////
  // in safeMode == 0 green LEDs are ON, waiting for initial off-balance condition
  ///////////////////////////////////////////////////////////////////////
  // the user must move the robot to an off-vertical value of >= 5 degrees
  if (safeMode == 0) {
    LED_AY = true; LED_BY = true; LED_CY = true;  LED_DY = true;
    if (AccVectTrk >= AccVectSafe) {
      if (OffBalance < 100) {OffBalance++;} // capture off-balance condition
      else {safeMode = 1;} // off-balance by more than 5 degrees so change safe modes
    } else {OffBalance = 0;}
  } else {
    ///////////////////////////////////////////////////////////////////////
    // safeMode > 0 so look for balance point and then go live
    ///////////////////////////////////////////////////////////////////////
    if (AccVectTrk <= AccVectTrgt) {
      // small vector size confirms 0.5 degree centre region
      LED_AB = HIGH; LED_AY = HIGH;
      LED_BB = HIGH; LED_BY = HIGH;
      LED_CB = HIGH; LED_CY = HIGH;
      LED_DB = HIGH; LED_DY = HIGH;
      GoActive++; if (GoActive >= 2) {
        ///////////////////////////////////////////////////////////////////////
        // robot moved to vertical position, so go active
        ///////////////////////////////////////////////////////////////////////
        setMainMode(3);                                   // switch modes and set safety flag
        angle_gyroX = angle_accX; pid_setpointX = 0.0;    // Load the accelerometer angle in the angle_gyro variable
        angle_gyroFX = angle_gyroX;                       // correct rate limited gyro angle
        self_balance_pid_setpointX = 0.0;                 // reset any previous set points
        angle_gyroY = angle_accY; pid_setpointY = 0.0;    // Load the accelerometer angle in the angle_gyro variable
        angle_gyroFY = angle_gyroY;                       // correct rate limited gyro angle
        self_balance_pid_setpointY = 0.0;                 // reset any previous set points
        GyroOnce = false;                                 // match gyro angle with accelerometer
        pid_i_memX = 0.0;                                 // reset integrators
        pid_i_memY = 0.0;                                 // reset integrators
        pid_error_lastX = 0.0;                            // reset d error memories
        pid_error_lastY = 0.0;                            // reset d error memories
        PID_TuneX = 0; PID_TuneY = 0;
        pid_output_trendX = 0; pid_output_trendY= 0;
        turnLEDsON(); LED_Override = 100; LEDBrightness = 8;
        self_balance_pid_incX = 0.0005; // values used to inc/dec balance setpoint (0.0015)
        self_balance_pid_incY = 0.0005; // values used to inc/dec balance setpoint (0.0015)
        PID_Output_Power = 50.0;                          // soft start from 50%
        startEn = true;
      }
    } else {
      // not yet at upright balance point, so give LED indication
      GoActive = 0; // reset the Go Active counter
      angle_gyroX = 0.0; angle_gyroY = 0.0; // lock gyros prior to balance point
      LEDCnt--; if (LEDCnt <= 5) {
        // LEDs are OFF for 50ms pulse
        if (LEDCnt < 1) {
          // count reached zero so reset based on current vector size
          LEDCnt = 10000 /((int)AccVectTrk + 1);
          LEDCnt = max(LEDCnt,11); LEDCnt = min(LEDCnt,99);
        }
      } else {
        // LEDs will be ON
        // light the LED which corresponds to the highest edge to indicate the
        // degree of out of balance
        if ((accelerometer_data_rawX_OA < 0) && (accelerometer_data_rawY_OA >= 0)) {
          // front left quadrant
          if (AccAbsAng >= 5.0269) {LED_AY = HIGH; LED_AB = HIGH;}
          else if (AccAbsAng >= 1.4766) {LED_AY = HIGH;}
          else if (AccAbsAng >= 0.6682) {LED_DB = HIGH; LED_AY = HIGH;}
          else if (AccAbsAng >= 0.1989) {LED_DB = HIGH;}
          else {LED_DB = HIGH; LED_DY = HIGH;}
        } else if ((accelerometer_data_rawX_OA < 0) && (accelerometer_data_rawY_OA < 0)) {
          // front right quadrant
          if (AccAbsAng >= 5.0269) {LED_AY = HIGH; LED_AB = HIGH;}
          else if (AccAbsAng >= 1.4766) {LED_AB = HIGH;}
          else if (AccAbsAng >= 0.6682) {LED_AB = HIGH; LED_BY = HIGH;}
          else if (AccAbsAng >= 0.1989) {LED_BY = HIGH;}
          else {LED_BB = HIGH; LED_BY = HIGH;}
        } else if ((accelerometer_data_rawX_OA >= 0) && (accelerometer_data_rawY_OA < 0)) {
          // rear right quadrant
          if (AccAbsAng >= 5.0269) {LED_CB = HIGH; LED_CY = HIGH;}
          else if (AccAbsAng >= 1.4766) {LED_CY = HIGH;}
          else if (AccAbsAng >= 0.6682) {LED_BB = HIGH; LED_CY = HIGH;}
          else if (AccAbsAng >= 0.1989) {LED_BB = HIGH;}
          else {LED_BB = HIGH; LED_BY = HIGH;}
        } else {
         // rear left quadrant
          if (AccAbsAng >= 5.0269) {LED_CB = HIGH; LED_CY = HIGH;}
          else if (AccAbsAng >= 1.4766) {LED_CB = HIGH;}
          else if (AccAbsAng >= 0.6682) {LED_CB = HIGH; LED_DY = HIGH;}
          else if (AccAbsAng >= 0.1989) {LED_DY = HIGH;}
          else {LED_DB = HIGH; LED_DY = HIGH;}
        }
      }
    }
  }
}

// --------------------------------------------------------------------------------

void MainMode_2() {
  ///////////////////////////////////////////////////////////////////////
  // Phantom MainMode 1
  ///////////////////////////////////////////////////////////////////////
  // mode setup is completed with only the PID P-gain value set, all other
  // gains are set to zero.
  // now switch back to MainMode 1
  MainMode = 1;
}

// --------------------------------------------------------------------------------

void MainMode_3() {
  ///////////////////////////////////////////////////////////////////////
  // Main self-balancing code
  ///////////////////////////////////////////////////////////////////////
  // bring up output power slowly to full on
  // then track PWM outputs with LEDs
  if (PID_Output_Power < 100.0) {PID_Output_Power += 0.5;}
}

// --------------------------------------------------------------------------------

void MainMode_4() {
  ///////////////////////////////////////////////////////////////////////
  // Upside down motor drive demo
  ///////////////////////////////////////////////////////////////////////
  // after an initial pause to allow the ball to be placed within the wheels
  // we start with circular motion
  if (Z_orientation >= 0) {setMainMode(0); return;} // exit if turned over
  
  // slow this process down if needed, without affecting the '99' delay timer
  // set MainSkipCnt = 1 or less to run on every cycle
  // set MainSkipCnt = 2 to run every second cycle (1/2 speed), etc
  if (ModeTask < 99) {
    MainSkip--; if (MainSkip > 0) {return;}
    MainSkip = MainSkipCnt; // reset the skip counter
  }
  
  
  switch(ModeTask) {
    case 0:
      // apply a 5 second delay before starting motor movements
      ModeCnt++; if (ModeCnt > 1250) {ModeTask++;}
      break;
    case 1:
      // initialise the variables for starting the run
      ModeSteer = 0; ModePID = 0.0;
      ModeCycle = 0; SetMainSkip(2); ModeTask++;
      break;
    case 2:
      // drive the motors with increasing steering only
      // from 0 to 255, this will rotate the ball at increasing speed
      DriveEn = true;
      driveMotors(0.0,0.0,ModeSteer);
      ModeSteer += 1;
      if (ModeSteer > 255) {
        // stay at this speed for 2 second then slow down & change direction
        ModeSteer = 254; ModeTaskNext = 3; ModeTask = 99; ModeCnt = 500;
      } break;
    case 3:
      // drive the motors with decreasing/reversing steering
      // from 254 to -255, this will rotate the ball at decreasing speed
      // then reverse its direction
      driveMotors(0.0,0.0,ModeSteer);
      if (ModeSteer == 0) {
        ModeCycle++;
        if (ModeCycle >= 2) {
          // pause for 2 seconds then switch task
          ModeTaskNext = 10; ModeTask = 99; ModeCnt = 500;
        }
      }
      ModeSteer -= 1;
      if (ModeSteer < -255) {
        // stay at this speed for 2 second then slow down & change direction
        ModeSteer = -254; ModeTaskNext = 2; ModeTask = 99; ModeCnt = 500;
      } break;

    // now we perform rolling movements
    case 10:
      // initialise the sequence
      ModePID = 0.5;    // counter used as the vector
      ModeCycle = 0; ModeTask++; break;
    case 11:
      // for each direction speed up to max = 255.0
      if (ModeCycle == 0) {driveMotors(ModePID,0.0,0);}
      if (ModeCycle == 1) {driveMotors(0.0,ModePID,0);}
      if (ModePID == 0.0) {
        ModeCycle++;
        // pause for 2 seconds then switch task
        if (ModeCycle == 2) {
          ModeTaskNext = 20; ModeTask = 99; ModeCnt = 500;
        }
      }
      ModePID += 0.5;
      if (ModePID > 255.0) {
        // run at full speed for 2 second
        ModePID >= 254.5; ModeTaskNext = 12; ModeTask =99; ModeCnt = 500;
      } break;
    case 12:
      // now slow down to zero and increment the angle by 120 degrees
      if (ModeCycle == 0) {driveMotors(ModePID,0.0,0);}
      if (ModeCycle == 1) {driveMotors(0.0,ModePID,0);}
      ModePID -= 0.5;
      if (ModePID < -255.0) {
        // run at full speed for 2 second
        ModePID = -254.5; ModeTaskNext = 11; ModeTask = 99; ModeCnt = 500;
      } break;

    // now we perform a rotating rolling movement
    case 20:
      // initialise the sequence
      ModePID = 0.5;    // counter used as the vector
      SetMainSkip(0); ModeCycle = 0; ModeTask++; break;
      break;
    case 21:
      // for each direction speed up to max = 255.0
      if (ModeCycle == 0) {driveMotors(ModePID,ModePID,0);}
      if (ModeCycle == 1) {driveMotors(ModePID,-ModePID,0);}
      if (ModePID == 0.0) {
        ModeCycle++;
        // pause for 2 seconds then switch task
        if (ModeCycle >= 2) {
          ModeTask = 0; ModeCnt = 750; SetMainSkip(0); // restart the sequence after 2 seconds
//          Serial.println("Exit 21");
          return;
        }
      }
      ModePID += 0.5;
      if (ModePID > 255.0) {
        // run at full speed for 2 second
        ModePID >= 254.5; ModeTaskNext = 22; ModeTask =99; ModeCnt = 500;
      } break;
    case 22:
      // now slow down to zero and increment the angle by 120 degrees
      if (ModeCycle == 0) {driveMotors(ModePID,ModePID,0);}
      if (ModeCycle == 1) {driveMotors(ModePID,-ModePID,0);}
      ModePID -= 0.5;
      if (ModePID < -255.0) {
        // run at full speed for 2 second
        ModePID = -254.5; ModeTaskNext = 21; ModeTask = 99; ModeCnt = 500;
      } break;
      
    case 99:
      // provides a delay of ModeCnt * 10ms before going to ModeTaskNext
      ModeCnt--; if (ModeCnt < 1) {ModeTask = ModeTaskNext;}
      break;
  }
}

// --------------------------------------------------------------------------------

void MainMode_Test() {
  ///////////////////////////////////////////////////////////////////////
  // A special test mode used for debugging
  ///////////////////////////////////////////////////////////////////////
  // after an initial pause to allow the ball to be placed within the wheels
  // we start with circular motion
  Z_orientation = -1; // over-ride the oriantation flag, so always upside down
  MainMode_4();       // run upside down mode continuously until reset
}

// --------------------------------------------------------------------------------
