

// ----------------------------------------------------------------------

void clearArrays() {
  // called to reset the contents of the arrays
  for (int zP = 0;zP < GraphDepth;zP++) {
    GraphVal0[zP] = 0; GraphVal1[zP] = 0; GraphVal2[zP] = 0; GraphVal3[zP] = 0;
  }
}

// ----------------------------------------------------------------------

void closeCOMPort() {
  // Close the COM port and empty the serial buffer
  if (comFlag > 0) {
    usbPort.clear(); usbPort.stop();
    comFlag = 0; msgCOM = "----"; drawFlag = 1;
    println("COM Port Closed");
  }
}

// ----------------------------------------------------------------------

void decDSlider() {
  // decrement 'D' gain slider value
  if (DsliderVal > DsliderVLL) {
    DsliderVal--;
    setDSlider();
    sendDgain(); // send the value to Arduino
    DsliderLast = DsliderVal;
  } DecDsliderDwn = true;
}

// ----------------------------------------------------------------------

void decISlider() {
  // decrement 'I' gain slider value
  if (IsliderVal > IsliderVLL) {
    IsliderVal--;
    setISlider();
    sendIgain(); // send the value to Arduino
    IsliderLast = IsliderVal;
  } DecIsliderDwn = true;
}

// ----------------------------------------------------------------------

void decSSlider() {
  // decrement 'S' gain slider value
  if (SsliderVal > SsliderVLL) {
    SsliderVal--;
    setSSlider();
    sendSgain(); // send the value to Arduino
    SsliderLast = SsliderVal;
  } DecSsliderDwn = true;
}

// ----------------------------------------------------------------------

void decPSlider() {
  // decrement 'P' gain slider value
  if (PsliderVal > PsliderVLL) {
    PsliderVal--;
    setPSlider();
    sendPgain(); // send the value to Arduino
    PsliderLast = PsliderVal;
  } DecPsliderDwn = true;
}

// ----------------------------------------------------------------------

void exit() {
  // program is closing so reset Arduino...
  usbPortWrite("!");
  delay(20); // allow 20ms for data to be sent
  super.exit();
}

// ----------------------------------------------------------------------

void getCOMPort() {
  // initialise serial comms if a port exists
  if (comFlag > 0) {
    // stops an existing connection if already made
    usbPort.clear(); usbPort.stop();
    comFlag = 0; msgCOM = "----"; delay(100);
  }
  comListLength = Serial.list().length;
  if (comListLength > 0) {
    comPnt++; if (comPnt >= comListLength) {comPnt = 0;}
    // test available serial port
    comName = Serial.list()[comPnt];
    println(comName);
    try {
      usbPort = new Serial(this, comName, 115200);
      //if (pingON > 0) {tick = 100;} // don't send ticks too early or board wont reset!
      usbPort.bufferUntil(LF);
      comFlag = 1; msgCOM = comName;
    } catch(Exception e) {
      comFlag = 0; msgCOM = "-Error-";
    }
  } else {
    // no serial port available
    comFlag = 0; msgCOM = "-NA-";
  } drawFlag = 1;
}

// ----------------------------------------------------------------------

void GraphButton() {
  // toggles Graph mode ON/OFF
  if (G_ON) {
    // return to PID windows
    G_ON = false; drawFlag = 1;
    usbPortWrite("RE.");  // send graph exit command
  } else {
    // initialise graphing window
    G_ON = true;      // switch to graph mode
    G_ReqCnt = 9;     // prime G_ReqCnt to send RPnn request
    GraphTitle = "";  // clear graph title
    GMax = 0.5;       // initial graph limits
    GMin = 0.5;       // initial graph limits
  }
}

// ----------------------------------------------------------------------

void GraphPressed() {
  // mouse has been clicked on graph window
  if ((mX < 26) && (mY < 26)) {toggleMouseON(); return;}
  
  if ((mX >= 1120) && (mX <= 1180)) {
    // check buttons
    
    // toggle graph/datya mode
    if ((mY >=  7) && (mY <= 40)) {if (G_Mode == 0) {G_Mode = 1;} else {G_Mode = 0;} return;}

    // check plot buttons
    if ((mY >= 49) && (mY <= 111)) {G_Plot0 = !G_Plot0; G_Draw = true; return;}
    if ((mY >= 111) && (mY <= 171)) {G_Plot1 = !G_Plot1; G_Draw = true; return;}
    if ((mY >= 171) && (mY <= 232)) {G_Plot2 = !G_Plot2; G_Draw = true; return;}
    if ((mY >= 232) && (mY <= 293)) {G_Plot3 = !G_Plot3; G_Draw = true; return;}
    
    // check map copy to clipboard button
    if ((mY >= 318) && (mY <= 379)) {copyGraphData(); return;}
    
    // check map scaling buttons
    if ((mY >= 397) && (mY <= 457)) {GraphPlus(); return;}
    if ((mY >= 457) && (mY <= 518)) {GraphMinus(); return;}
    if ((mY >= 518) && (mY <= 580)) {GraphAuto(); return;}
    
    // check GraphDepth field
    if ((mY >= 604) && (mY <= 628)) {SetGraphDepth(); return;}
    
    // check RPnn field
    if ((mY >= 655) && (mY <= 680)) {SetRPnn(); G_First = true; return;}
    
    // check Pause/Play button
    if ((mY >= 689) && (mY <= 750)) {G_Pause = !G_Pause; return;}
  }
  
  // check right-hand side of buttons
  if ((mX > 1180) && (mY >= 49) && (mY <= 293)) {
     G_Plot0 = true; G_Plot1 = true; G_Plot2 = true; G_Plot3 = true;
     G_Draw = true; return;
  }
  // default is clicked to exit
  GraphButton();
}

// ----------------------------------------------------------------------

void incDSlider() {
  // increment 'D' slider value
  DsliderVal++;
  if (DsliderVal > DsliderVUL) {DsliderVUL = DsliderVal;}
  setDSlider();
  sendDgain(); // send the value to Arduino
  DsliderLast = DsliderVal;
  IncDsliderDwn = true;
}

// ----------------------------------------------------------------------

void incISlider() {
  // increment 'I' slider value
  IsliderVal++;
  if (IsliderVal > IsliderVUL) {IsliderVUL = IsliderVal;}
  setISlider();
  sendIgain(); // send the value to Arduino
  IsliderLast = IsliderVal;
  IncIsliderDwn = true;
}

// ----------------------------------------------------------------------

void incPSlider() {
  // increment 'P' slider value
  PsliderVal++;
  if (PsliderVal > PsliderVUL) {PsliderVUL = PsliderVal;}
  setPSlider();
  sendPgain(); // send the value to Arduino
  PsliderLast = PsliderVal;
  IncPsliderDwn = true;
}

// ----------------------------------------------------------------------

void incSSlider() {
  // increment 'S' slider value
  SsliderVal++;
  if (SsliderVal > SsliderVUL) {SsliderVUL = SsliderVal;}
  setSSlider();
  sendSgain(); // send the value to Arduino
  SsliderLast = SsliderVal;
  IncSsliderDwn = true;
}

// ----------------------------------------------------------------------

void loadScrns() {
  // load background images
  imgPID = loadImage("PID Controller.png");
  imgGraph = loadImage("Graph.png");
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void MemButton() {
  // user clicked on Mem button
  if (MemDown) {
    // button is being released
    MemDown = false;
    if ((millis() - MemMillis) > 1000) {
      // store PID values
      Mem_D = DsliderVal;
      Mem_I = IsliderVal;
      Mem_P = PsliderVal;
      Mem_S = SsliderVal;
      M_ON = true; // indicate values are being stored/recovered
      println("PID --> Memory");
    } else {
      // use stored PID values if they exist
      if (M_ON) {
        DsliderVal = Mem_D; setDSlider(); sendDgain();
        IsliderVal = Mem_I; setISlider(); sendIgain();
        PsliderVal = Mem_P; setPSlider(); sendPgain();
        SsliderVal = Mem_S; setSSlider(); sendSgain();
        println("Memory --> PID");
      }
    }
  } else {
    // Mem button pressed so store time of click
    MemMillis = millis(); MemDown = true;
  } drawFlag = 1;
}

// ----------------------------------------------------------------------

void PIDSetpointJolt() {
  // toggle the setpoint joly modes
  PID_SPJ++; if (PID_SPJ > 3) {PID_SPJ = 0;}
  if (PID_SPJ < 1) {msgTx = "SS0.";}
  if (PID_SPJ == 1) {msgTx = "SS1.";}
  if (PID_SPJ == 2) {msgTx = "SS2.";}
  if (PID_SPJ == 3) {msgTx = "SS3.";}
  usbPortWrite(msgTx); drawFlag = 1;
}

// ----------------------------------------------------------------------

void readDSlider() {
  // determine D-slider position and send value
  JoyDval = mX; JoyDval = max(JoyDval, JoyX0);
  JoyDval = min(JoyDval, JoyX1);
  // normal single channel mode
  DsliderVal = ((JoyDval - JoyX0) * DsliderVUL) / (JoyX1 - JoyX0);
  DsliderVal = max(DsliderVal, DsliderVLL);
  DsliderVal = min(DsliderVal, DsliderVUL);
  if (DsliderLast != DsliderVal) {
    sendDgain(); // send the value to Arduino
    DsliderLast = DsliderVal;
  }
}

// ----------------------------------------------------------------------

void readISlider() {
  // determine I-slider position and send value
  JoyIval = mX; JoyIval = max(JoyIval, JoyX0);
  JoyIval = min(JoyIval, JoyX1);
  // normal single channel mode
  IsliderVal = ((JoyIval - JoyX0) * IsliderVUL) / (JoyX1 - JoyX0);
  IsliderVal = max(IsliderVal, IsliderVLL);
  IsliderVal = min(IsliderVal, IsliderVUL);
  if (IsliderLast != IsliderVal) {
    sendIgain(); // send the value to Arduino
    IsliderLast = IsliderVal;
  }
}

// ----------------------------------------------------------------------

void readPSlider() {
  // determine P-slider position and send value
  JoyPval = mX; JoyPval = max(JoyPval, JoyX0);
  JoyPval = min(JoyPval, JoyX1);
  // normal single channel mode
  PsliderVal = ((JoyPval - JoyX0) * PsliderVUL) / (JoyX1 - JoyX0);
  PsliderVal = max(PsliderVal, PsliderVLL);
  PsliderVal = min(PsliderVal, PsliderVUL);
  if (PsliderLast != PsliderVal) {
    sendPgain(); // send the value to Arduino
    PsliderLast = PsliderVal;
  }
}

// ----------------------------------------------------------------------

void readSSlider() {
  // determine S-slider position and send value
  JoySval = mX; JoySval = max(JoySval, JoyX0);
  JoySval = min(JoySval, JoyX1);
  // normal single channel mode
  SsliderVal = ((JoySval - JoyX0) * SsliderVUL) / (JoyX1 - JoyX0);
  SsliderVal = max(SsliderVal, SsliderVLL);
  SsliderVal = min(SsliderVal, SsliderVUL);
  if (SsliderLast != SsliderVal) {
    sendSgain(); // send the value to Arduino
    SsliderLast = SsliderVal;
  }
}

// ----------------------------------------------------------------------

void sendDot() {
  // send a '.' dot character to the Arduino to get it to report PID gains
  msgTx = "."; usbPortWrite(msgTx);
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void sendRESET() {
  // reset defaults and send RESET msg to Arduino
  msgTx = "!"; usbPortWrite(msgTx);
  resetFlags(); drawFlag = 1;
}

// ----------------------------------------------------------------------

void sendDgain() {
  // send the 'D' gain value message to Arduino
  msgTx = "DV" + str(DsliderVal) + ".";
  msgDgain = msgTx; msgDnew = true;
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void sendIgain() {
  // send the 'I' gain value message to Arduino
  msgTx = "IV" + str(IsliderVal) + ".";
  msgIgain = msgTx; msgInew = true;
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void sendPgain() {
  // send the 'P' gain value message to Arduino
  msgTx = "PV" + str(PsliderVal) + ".";
  msgPgain = msgTx; msgPnew = true;
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void sendSgain() {
  // send the 'S' gain value message to Arduino
  msgTx = "SG" + str(SsliderVal) + ".";
  msgSgain = msgTx; msgSnew = true;
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void setDSlider() {
  // set the 'D' gain slider position based on DsliderVal value
  DsliderVal = min(DsliderVal, DsliderVUL);
  DsliderVal = max(DsliderVal, DsliderVLL);
  JoyDval = JoyX0 + ((DsliderVal * (JoyX1 - JoyX0))/(DsliderVUL - DsliderVLL)); 
  JoyDval = max(JoyX0, JoyDval); // prevent it going too low
  JoyDval = min(JoyX1, JoyDval); // prevent it going too high
  DsliderLast = DsliderVal;
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void setISlider() {
  // set the 'I' gain slider position based on IsliderVal value
  IsliderVal = min(IsliderVal,IsliderVUL);
  IsliderVal = max(IsliderVal, IsliderVLL);
  JoyIval = JoyX0 + ((IsliderVal * (JoyX1 - JoyX0))/(IsliderVUL - IsliderVLL)); 
  JoyIval = max(JoyX0, JoyIval); // prevent it going too low
  JoyIval = min(JoyX1, JoyIval); // prevent it going too high
  IsliderLast = IsliderVal;
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void setPSlider() {
  // set the 'P' gain slider position based on PsliderVal value
  PsliderVal = min(PsliderVal, PsliderVUL);
  PsliderVal = max(PsliderVal, PsliderVLL);
  JoyPval = JoyX0 + (((PsliderVal - PsliderVLL) * (JoyX1 - JoyX0))/(PsliderVUL - PsliderVLL)); 
  JoyPval = max(JoyX0, JoyPval); // prevent it going too low
  JoyPval = min(JoyX1, JoyPval); // prevent it going too high
  PsliderLast = PsliderVal;
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void setSSlider() {
  // set the 'S' gain slider position based on SsliderVal value
  SsliderVal = min(SsliderVal, SsliderVUL);
  SsliderVal = max(SsliderVal, SsliderVLL);
  JoySval = JoyX0 + (((SsliderVal - SsliderVLL) * (JoyX1 - JoyX0))/(SsliderVUL - SsliderVLL)); 
  JoySval = max(JoyX0, JoySval); // prevent it going too low
  JoySval = min(JoyX1, JoySval); // prevent it going too high
  SsliderLast = SsliderVal;
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void toggleD() {
  // toggle the PID D factor ON/OFF
  if (D_ON) {D_ON = false; msgTx = "D0.";}
  else {D_ON = true; msgTx = "D1.";}
  usbPortWrite(msgTx); drawFlag = 1;
}

// ----------------------------------------------------------------------

void toggleI() {
  // toggle the PID I factor ON/OFF
  if (I_ON) {I_ON = false; msgTx = "I0.";}
  else {I_ON = true; msgTx = "I1.";}
  usbPortWrite(msgTx); drawFlag = 1;
}

//-------------------------------------------------------------------------

void toggleMouseON() {
  // toggle mouse ON/OFF flag
  if (mON) {
    mON = false;
  } else {
    mON = true; mXL = -1; mYL = -1;
  } drawFlag = 1;
}

// ----------------------------------------------------------------------

void usbPortWrite(String zmsg) {
  // called to send text over the serial port
  if (comFlag > 0) {usbPort.write(zmsg);}
}
// ----------------------------------------------------------------------
