// ######################################################################
//
//  PID Controller 4x4
//  Released:  10/04/2022    By: Tony Swanwick
//
// ######################################################################
/*
  This bespoke code communicates with the Arduino NANO in the Ball BalanceBot
  to control the PID controller variables when in DEBUG mode and connected
  over the WiFi link. In effect you can experiment with the variables to
  see what affect they have on the robots balance and stability.
  
  This version includes:
  * PID slider pots
  * Graph window
  * Copy graph data to clipboard
  
*/

// declare libraries
import processing.serial.*;
Serial usbPort;

String Released$ = "10/04/2022";

// declare Graph variables
boolean G_Auto = true;    // true if graph is in auto-scaling mode, default
boolean G_Draw = false;
boolean G_First = false;  // sets plot variables on first line of data
String GFormat;           // used for formatting floating point numbers
float GMax,GMax0,GMax1,GMax2,GMax3; // graph waveform max values
float GMin,GMin0,GMin1,GMin2,GMin3; // graph waveform min values
int G_Mode = 0;           // 0 = graphing, 1 = data listing
boolean G_ON = false;     // true when Graph mode window is displayed
boolean G_Pause = false;  // set true to pause graphing function
boolean G_Plot0 = true;   // line plotting flags
boolean G_Plot1 = true;   // line plotting flags
boolean G_Plot2 = true;   // line plotting flags
boolean G_Plot3 = true;   // line plotting flags
int G_Pp,G_Ps;            // graph point position pointers
int G_Py0,G_Py1,G_Py2,G_Py3; // previous Y-axis values to plot
int G_PxInc = 4;          // graph horizontal increment
int GraphDepth = 500;     // number of points in the graph
int GraphMargin = 100;    // number of pixels indented from the left
int GraphPnt = 0;         // a pointer used in graphing
int G_ReqCnt = 3;         // data request counter timer
int GraphRange = 0;       // number of values stored in arrays
Boolean GraphSkip = true; // ignores the first reading of data
String GraphTitle = "";   // received title
String GraphTitle0 = "";  // title element 0
String GraphTitle1 = "";  // title element 1
String GraphTitle2 = "";  // title element 2
String GraphTitle3 = "";  // title element 3
float [] GraphVal0 = new float[GraphDepth];
float [] GraphVal1 = new float[GraphDepth];
float [] GraphVal2 = new float[GraphDepth];
float [] GraphVal3 = new float[GraphDepth];
float G_Scale = 10.0; // graph manual scaling factor
int GTCX0,GTCX1,GTCX2,GTCX3;  // title column X values
int G_Vy0,G_Vy1,G_Vy2,G_Vy3;  // Y-axis values to plot
float G_WinMax = 750.0;   // max vertical pixel value
float G_WinMin = 50.0;    // min vertical pixel value
int GX = 0;               // mouse X relative to graph area
float GY = 0;             // mouse Y relative to graph area
int NumTraces;            // number of values received and therefore traces
long RPms = millis();     // timer used in 40ms R. requests
int RPnn = 0;             // report flag, selects data to be sent from Arduino
int RPnnMax = 6;          // highest RPnn number used in data responses

// declare variables
int comFlag = 0;          // =1 if COM serial available
int comListLength = 0;    // number of items in COM connection list
String comName = "";      // serial port COM name
int comPnt = 0;           // COM port list pointer
String data = "";
boolean D_ON;             // 'D' flag; true = ON, false = OFF
boolean DecDsliderDwn = false; // false when mouse is clicked on slider
boolean DecIsliderDwn = false; // false when mouse is clicked on slider
boolean DecPsliderDwn = false; // false when mouse is clicked on slider
boolean DecSsliderDwn = false; // false when mouse is clicked on slider
int drawFlag = 1;         // = 1 to force screen redraw
boolean DsliderDwn = false; // false when mouse is clicked on slider
int DsliderLast;          // previous slider value
int DsliderVal;           // slider value 0 - 400
int DsliderVLL = 0;       // 'D' gain lower limit
int DsliderVUL = 500;     // 'D' gain upper limit/10; ie 500 = 50.0
float zF;                 // any temp float value
String Help$ = "";        // help message
boolean IncDsliderDwn = false; // false when mouse is clicked on slider
boolean IncIsliderDwn = false; // false when mouse is clicked on slider
boolean IncPsliderDwn = false; // false when mouse is clicked on slider
boolean IncSsliderDwn = false; // false when mouse is clicked on slider
boolean I_ON;             // 'I' flag; true = ON, false = OFF
PImage imgPID;            // image loaded as background
PImage imgGraph;          // image loaded as background
int interval = 30;        // main loop interval in milliseconds
boolean IsliderDwn = false; // true when mouse is clicked on slider
int IsliderLast;          // previous slider value
int IsliderVal;           // slider value 0 - 500, divided by 100 at Arduino
int IsliderVLL = 0;       // 'I' gain lower limit
int IsliderVUL = 500;    // 'I' gain upper limit/100; ie. 500 = 5.00
int JoyDval;              // 'D' gain joystick slider x-value
int JoyIval;              // 'I' gain joystick slider x-value
int JoyPval;              // 'P' gain joystick slider value
int JoySval;              // 'S' gain joystick slider value
int JoyX0 = 166;          // x-coordinate of joystick slider left end
int JoyX1 = 616;          // x-coordinate of joystick slider right end
int JoyXM = 390;          // centre of slider region
int LF = 10;              // ASCII line feed
char Ln = 10;             // ASCII line feed
boolean mDown,mON;        // mouse flag and last x,y
int Mem_D = 0;            // stored DsliderVal
int Mem_I = 0;            // stored IsliderVal
int Mem_P = 0;            // stored PsliderVal
int Mem_S = 0;            // stored SsliderVal
boolean MemDown;          // Mem button clicked = true
int MemMillis;            // Mem button timer in milliseconds
boolean M_ON = false;     // 'M' flag; true = ON, false = OFF
int mB,mTO,mX,mXL,mY,mYL; // mouse flag and last x,y
String msgCh;             // channel message sent to Arduino
int msgCnt = 0;           // counter used in message error checking
String msgCOM = "None";   // COM port name
String msgDgain = "";     // latest D gain message
boolean msgDnew = false;  // == true if there is a msgDgain to send
String msgIgain = "";     // latest I gain message
boolean msgInew = false;  // == true if there is a msgIgain to send
String msgPgain = "";     // latest P gain message
boolean msgPnew = false;  // == true if there is a msgPgain to send
int msgPnt = 0;           // pointer to message buffer to be sent
String msgRx = "";        // message received from Arduino
String msgRxLast = "";    // previous message received from Arduino
String msgSgain = "";     // latest S gain message
boolean msgSnew = false;  // == true if there is a msgSgain to send
String msgTx = "";        // message sent to Arduino
int nextMillis = 0;       // loop end time in millis
int PID_SPJ = 0;          // PID setpoint jolt mode
boolean PsliderDwn = false; // true when mouse is clicked on slider
int PsliderLast;          // previous slider value
int PsliderVal;           // slider value 50 - 550, divide by 10 at Arduinio
int PsliderVLL = 0;       // 'P' gain lower limit
int PsliderVUL = 800;    // 'P' gain upper limit/10; ie. 500 = 50.0
boolean SsliderDwn = false; // true when mouse is clicked on slider
int SsliderLast;          // previous slider value
int SsliderVal;           // slider value 0 - 500, divided by 100 at Arduino
int SsliderVLL = 0;       // 'S' gain lower limit
int SsliderVUL = 500;     // 'S' gain upper limit/10000; ie. 500 = 0.05
int TB = 0;               // text Blue colour component
int TG = 0;               // text Green colour component
int TR = 0;               // text Red colour component
String usbLast = "";      // last message rejected by the USB port
int usbPNotReady = 0;     // counter used to prevent port over-run at 9600 baud
String zTxt;              // any temp text value

void setup() {
  // first code to run
  mON = false; // mouse X,Y is OFF initially
  resetFlags();
  InitGraphDraw();
  size(1200,800);  // screen size must match background bitmap
  loadScrns(); // load the default screen
  frameRate(100); // set max screen framerate to 100fps
  getCOMPort();
  // allow Arduino time to boot up
  delay(100);
  println("Ready!");
  nextMillis = millis() + interval;
}

void draw() {
  // main loop. Comes here about 100 times a second
  if (mDown) {mTO--; if (mTO < 1) {mouseRepeat(); mTO = 6;}}
  if (mON) {drawMouseXY();}
  if (G_ON) {
    // if in Graph mode request data from the Ball BalanceBot
    // this occurs approximately every 40ms
    if (G_ReqCnt > 0) {G_ReqCnt--;}
    if (G_ReqCnt == 6) {
      // changing RPnn request at the start of the graphing process
      msgTx = "RP" + RPnn + "."; usbPortWrite(msgTx);
      InitGraphDraw();
      RPms = millis();
    }
    else if (G_ReqCnt < 1) {
      if ((millis() - RPms) >= 40) {
        //println(millis() - RPms);
        RPms = millis();
        if (!G_Pause) {
          // request more data from the robot
          usbPortWrite("R.");
        } else {
          // in 'Pause' so draw graph + data under mouse position
          G_Draw = true;
        }
      }
    }
  } else {
    // normal PID Controller mode
    if (drawFlag > 0) {
      background(imgPID); // define background RGB colour
      drawSliderP(); drawSliderI(); drawSliderD(); drawSsliderVal();
      drawPsliderVal(); drawIsliderVal(); drawDsliderVal(); drawSliderS();
      drawCOMMsg(); drawRxMsg(); drawTxMsg(); drawHelpMsg();
      drawLamps();
      if (mON) {drawMouseXY();}
      drawGraphMode();
      drawFlag = 0;
    }
    
    // to avoid overloading the WiFi link by rapid slider movement the values are
    // sent from buffers which hold the latest values, by cycling through these we
    // naturally introduce a delay in sending updates
    switch(msgPnt) {
      case 0: if (msgPnew) {usbPortWrite(msgPgain); msgPnew= false;} break;
      case 1: if (msgInew) {usbPortWrite(msgIgain); msgInew = false;} break;
      case 2: if (msgDnew) {usbPortWrite(msgDgain); msgDnew = false;} break;
      case 3: if (msgSnew) {usbPortWrite(msgSgain); msgSnew = false;} break;
    }
    msgPnt++; if (msgPnt > 3) {msgPnt = 0;}
  }
  // check for Mem button time-out
  if (MemDown) {
    if ((millis() - MemMillis) > 1000) {
      fill(100,255,0); ellipse(362,80,21,21); // Mem button lamp = green
    }
  }
  if (G_Draw) {
    // Graph draw flag has been set by received data
    if (G_Mode == 0) {drawAxis(); drawGraph();}
    else {drawData();}
    drawGraphTitle();
    drawGraphMode();
    G_Draw = false;
  }
}

// ----------------------------------------------------------------------

void keyPressed() {
  // handles keyboard entry
  if (keyCode == 32) {
    if (G_ON) {G_Pause = !G_Pause;}
  }
}

// ----------------------------------------------------------------------

void mouseDragged() {
  // mouse has been clicked and dragged on window
  if (G_ON) {return;}
  mX = mouseX; mY = mouseY; mDown = true;
  if (PsliderDwn) {
    // ignore mY if it goes outside of slider region
    if ((mX >=150) && (mX <=650)) {
      // mouse in P-slider region
      readPSlider(); drawFlag = 1;
    }
  }
  if (IsliderDwn) {
    // ignore mY if it goes outside of slider region
    if ((mX >=150) && (mX <=650)) {
      // mouse in I-slider region
      readISlider(); drawFlag = 1;
    }
  }
  if (DsliderDwn) {
    // ignore mY if it goes outside of slider region
    if ((mX >=150) && (mX <=650)) {
      // mouse in D-slider region
      readDSlider(); drawFlag = 1;
    }
  }
  if (SsliderDwn) {
    // ignore mY if it goes outside of slider region
    if ((mX >=150) && (mX <=650)) {
      // mouse in S-slider region
      readSSlider(); drawFlag = 1;
    }
  }
}

// ----------------------------------------------------------------------

void mouseMoved() {
  // called whenever the mouse moves
  // used for help message system
  mX = mouseX; mY = mouseY;
  if (G_ON) {return;}
  
  Help_Msg();
  if (mON) {drawFlag = 1;}
}

// ----------------------------------------------------------------------

void mousePressed() {
  // mouse has been clicked on window
  mX = mouseX; mY = mouseY; mB = mouseButton;
  if (G_ON) {GraphPressed(); return;}
  
  mDown = true; mTO = 30; // 30ms delay on mouse repeat
  if ((mX < 26) && (mY < 26)) {toggleMouseON();}

  // 'P' gain slider region?
  if ((mY >= 159) && (mY <= 196)) {
         if ((mX >=127) && (mX <=141)) {decPSlider();}
    else if ((mX >=155) && (mX <=635)) {
      // mouse in P-slider region
      PsliderDwn = true; readPSlider(); drawFlag = 1;
    }
    else if ((mX >=639) && (mX <=654)) {incPSlider();}
    else if ((mX >=674) && (mX <=764)) {sendPgain();}
  }

  // 'I' gain slider region?
  if ((mY >= 212) && (mY <= 249)) {
         if ((mX >=127) && (mX <=141)) {decISlider();}
    else if ((mX >=639) && (mX <=654)) {incISlider();}
    else if ((mX >=155) && (mX <=635)) {
      // mouse in I-slider region
      IsliderDwn = true; readISlider(); drawFlag = 1;
    }
    else if ((mX >=674) && (mX <=764)) {sendIgain();}
  }

  // 'D' gain slider region?
  if ((mY >= 266) && (mY <= 303)) {
         if ((mX >=127) && (mX <=141)) {decDSlider();}
    else if ((mX >=155) && (mX <=635)) {
      // mouse in D-slider region
      DsliderDwn = true; readDSlider(); drawFlag = 1;
    }
    else if ((mX >=639) && (mX <=654)) {incDSlider();}
    else if ((mX >=674) && (mX <=764)) {sendDgain();}
  }

  // 'S' gain slider region?
  if ((mY >= 322) && (mY <= 363)) {
         if ((mX >=127) && (mX <=141)) {decSSlider();}
    else if ((mX >=155) && (mX <=635)) {
      // mouse in D-slider region
      SsliderDwn = true; readSSlider(); drawFlag = 1;
    }
    else if ((mX >=639) && (mX <=654)) {incSSlider();}
    else if ((mX >=674) && (mX <=764)) {sendSgain();}
  }

  // Buttons
  if ((mY >= 64) && (mY <= 143)) {
    // RESET button
    if ((mX >= 36) && (mX <= 114)) {sendRESET();}
  
    // D button
    if ((mX >= 218) && (mX <= 297)) {toggleD();}
  
    // I button
    if ((mX >= 136) && (mX <= 214)) {toggleI();}
  
    // Mem button
    if ((mX >= 300) && (mX <= 379)) {MemButton();}
  
    // PID setpoint jolt button
    if ((mX >= 380) && (mX <= 460)) {PIDSetpointJolt();}
  }
  
  // clicked on COM port?
  if ((mX >= 88) && (mX <= 175) && (mY >= 384) && (mY <= 407)) {
    if (mouseButton == LEFT) {getCOMPort();}
    else if (mouseButton == RIGHT) {closeCOMPort();}
  }

  // Graph area
  if ((mX >= 800) || (mY >= 408)) {GraphButton();}
  
  // clicked on Tx field?
  if ((mX >= 635) && (mX <= 763) && (mY >= 384) && (mY <= 407)) {sendDot();}

}

// ----------------------------------------------------------------------

void mouseReleased() {
  // mouse released from click and drag operation
  if (G_ON) {return;}
  
  mX = mouseX; mY = mouseY;
  mDown = false;
  DecDsliderDwn = false; DecIsliderDwn = false; DecPsliderDwn = false;
  DecSsliderDwn = false;
  IncDsliderDwn = false; IncIsliderDwn = false; IncPsliderDwn = false;
  IncSsliderDwn = false;
  DsliderDwn = false; IsliderDwn = false; // clear slider flags
  PsliderDwn = false; SsliderDwn = false; // clear slider flags
  if (MemDown) {MemButton();}
}

// ----------------------------------------------------------------------

void mouseRepeat() {
  // called if mouse still down after a short pause period
  mX = mouseX; mY = mouseY;
  if (G_ON) {return;}
  if (DecPsliderDwn) {
    if ((mX >=127) && (mX <=141) && (mY >= 159) && (mY <= 196)) {decPSlider();}
  }
  if (DecIsliderDwn) {
    if ((mX >=127) && (mX <=141) && (mY >= 212) && (mY <= 249)) {decISlider();}
  }
  if (DecDsliderDwn) {
    if ((mX >=127) && (mX <=141) && (mY >= 266) && (mY <= 303)) {decDSlider();}
  }
  if (DecSsliderDwn) {
    if ((mX >=127) && (mX <=141) && (mY >= 322) && (mY <= 363)) {decSSlider();}
  }
  if (IncPsliderDwn) {
    if ((mX >=639) && (mX <=654) && (mY >= 159) && (mY <= 196)) {incPSlider();}
  }
  if (IncIsliderDwn) {
    if ((mX >=639) && (mX <=654) && (mY >= 212) && (mY <= 249)) {incISlider();}
  }
  if (IncDsliderDwn) {
    if ((mX >=639) && (mX <=654) && (mY >= 266) && (mY <= 303)) {incDSlider();}
  }
  if (IncSsliderDwn) {
    if ((mX >=639) && (mX <=654) && (mY >= 322) && (mY <= 363)) {incSSlider();}
  }
}

// ----------------------------------------------------------------------

void resetFlags() {
  // set flags to match RESET condition
  D_ON = true; // 'D' flag; true = ON, false = OFF
  DsliderVal = 0; // slider value 0 to 400
  G_PxInc = 4;  // graph horizontal increment
  GraphDepth = 250;  // number of points in the graph
  I_ON = true; // 'D' flag; true = ON, false = OFF
  IsliderVal = 0; // slider value 0 to 500
  JoyPval = JoyX0; // P joystick slider x-value
  JoyIval = JoyX0; // I joystick slider x-value
  JoyDval = JoyX0; // D joystick slider x-value
  JoySval = JoyX0; // S joystick slider x-value
  MemDown = false; // Mem button clicked = true
  MemMillis = 0;  // Mem button timer in milliseconds
  msgCh = "---"; // channel message sent to Arduino
  mTO = 0; // mouse down timeout
  PsliderVal = 0; // slider value 50 to 550
  SsliderVal = 0; // slider value 0 to 500
}

// ----------------------------------------------------------------------

void serialEvent(Serial usbPort) {
  // respond to a serial data received event
  data = usbPort.readStringUntil(LF);
  msgRx = trim(data); data = "";
  int CP;  // character pointer
  if (msgRx.length() < 1) {return;}
  
  if (!G_ON) {
    // not in Graph mode, so expect low rate mixed data
    println(msgRx);
    // look for PID values if string is long enough
    // format is: PID=p_gain, i_gain, d_gain
    if (msgRx.length() >= 15) {
      String[] m = match(msgRx,"PID=");
      if (m != null) {
        println("Decoding...");
        // remove the 'PID=' text
        msgRx = msgRx.substring(1 + msgRx.indexOf("="));
        
        // extract the first number, p gain
        CP = msgRx.indexOf(",");
        if (CP < 1) {return;}
        String Pgain = msgRx.substring(0,CP);
        println(Pgain);
        PsliderVal = int(10 * float(Pgain)); setPSlider();
        
        // extract the 2nd number, I gain
        msgRx = trim(msgRx.substring(1 + CP));
        CP = msgRx.indexOf(",");
        if (CP < 1) {return;}
        String Igain = trim(msgRx.substring(0,CP));
        println(Igain);
        if (Igain.indexOf("--.--") < 0) {
          // string does not indicate I gain is off
          IsliderVal = int(100 * float(Igain)); setISlider();
        } else {I_ON = false;}

        // extract the 3rd number, D gain
        msgRx = trim(msgRx.substring(1 + CP));
        CP = msgRx.indexOf(",");
        if (CP < 1) {return;}
        String Dgain = trim(msgRx.substring(0,CP));
        println(Dgain);
        if (Dgain.indexOf("--.--") < 0) {
          // string does not indicate D gain is off
          DsliderVal = int(10.0 * float(Dgain)); setDSlider();
        } else {D_ON = false;}
        
        // extract the last number, S gain
        String Sgain = trim(msgRx.substring(1 + CP));
        println(Sgain);
        SsliderVal = int(10000.0 * float(Sgain)); setSSlider();
        
        // put it back together for the Rx field
        msgRx = Pgain + "," + Igain + "," + Dgain + "," + Sgain;
      }
    }
  } else {
    // in Graph mode so expect high rate CSV data values
    // the Ball BalanceBot sends two matching consequitive value groups
    // two consequitive value groups must therefore match to be recognised
    //println(msgRx);
    if (msgRx.substring(0,1).equals("#")) {
      GraphTitle = msgRx.substring(1); msgRx = ""; msgRxLast = "";
      GraphTitle0 = ""; GraphTitle1 = ""; GraphTitle2 = ""; GraphTitle3 = "";

      // introduce spaces after the commas
      CP = GraphTitle.indexOf(",");
      if (CP >= 0) {
        // grab the title elements, up to 4 in total
        GraphTitle0 = GraphTitle.substring(0,CP);  GraphTitle = GraphTitle.substring(CP + 1);
        CP = GraphTitle.indexOf(",");
        if (CP >= 0) {
          GraphTitle1 = GraphTitle.substring(0,CP);  GraphTitle = GraphTitle.substring(CP + 1);
          CP = GraphTitle.indexOf(",");
          if (CP >= 0) {
            GraphTitle2 = GraphTitle.substring(0,CP);  GraphTitle3 = GraphTitle.substring(CP + 1);
          } else {GraphTitle2 = GraphTitle;}
        } else {GraphTitle1 = GraphTitle;}
      } else {GraphTitle0 = GraphTitle;}
    }
    else {
      msgCnt++;
      if (msgRx.equals(msgRxLast)) {
        // two consequitive messages match so use this message
        if (GraphSkip) {GMin = 0.0; GMax = 0.1; GraphSkip = false; return;}
        //println(msgRx);
        //G_ReqCnt = 0; // set flag for more messages to be sent
        float [] values = float(split(msgRx,','));
        NumTraces = values.length;
        GraphVal0[GraphPnt] = values[0];
        GraphVal1[GraphPnt] = 0.0;
        if (NumTraces > 1) {GraphVal1[GraphPnt] = values[1];}
        GraphVal2[GraphPnt] = 0.0;
        if (NumTraces > 2) {GraphVal2[GraphPnt] = values[2];}
        GraphVal3[GraphPnt] = 0.0;
        if (NumTraces > 3) {GraphVal3[GraphPnt] = values[3];}
        GraphPnt++; if (GraphPnt >= GraphDepth) {GraphPnt = 0;}
        if (GraphRange < GraphDepth) {GraphRange++;}
        // determine autorange values
        GraphMinMax();
        G_Draw = true;
        if (G_First) {
          // first line of data so determine which lines will be displayed
          G_First = false; G_Plot0 = true;
          if (NumTraces > 1) {G_Plot1 = true;}
          if (NumTraces > 2) {G_Plot2 = true;}
          if (NumTraces > 3) {G_Plot3 = true;}
        }
        msgRx = ""; // clear the message, otherwise the graph will speed up
        msgCnt = 0;  // reset the message counter
      } else {
        // no match, if it is the first of a pair then ignore it, otherwise assume
        // this is an error and report it
        //if (msgCnt > 1) {println(msgRx);}
      }
      msgRxLast = msgRx; // track the messages to look for a match
    }
  }
  drawFlag = 1;
}

// ----------------------------------------------------------------------
